import types
from collections.abc import Callable, Sequence
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Any

from sybil import Sybil, Document
from sybil.typing import Parser, Lexer, LexemeMapping


def check_sybil(sybil: Sybil, text: str) -> Document:
    """
    Run the supplied text through the supplied Sybil instance and evaluate the single
    example it contains.
    """
    with NamedTemporaryFile() as temp:
        temp.write(text.encode())
        temp.flush()
        document = sybil.parse(Path(temp.name))
    examples = list(document.examples())
    assert len(examples) == 1, f'Expected exactly one example, got: {examples}'
    examples[0].evaluate()
    return document


def check_parser(parser: Parser, text: str) -> Document:
    """
    Run the supplied text through the supplied parser and evaluate the single
    example it contains.

    This is for testing :data:`~sybil.typing.Parser` implementations.
    """
    sybil = Sybil(parsers=[parser], pattern='*')
    return check_sybil(sybil, text)


def check_lexer(
    lexer: Lexer, source_text: str, expected_text: str, expected_lexemes: LexemeMapping
) -> None:
    """
    Run the supplied text through the supplied lexer make sure it lexes a single
    region and captures the expected text and lexemes.

    This is for testing :data:`~sybil.typing.Lexer` implementations.
    """
    document = Document(source_text, 'sample.txt')
    regions = list(lexer(document))
    assert len(regions) == 1, f'Expected exactly one region, got: {regions}'
    region = regions[0]
    assert expected_text == document.text[region.start : region.end]
    assert region.lexemes == expected_lexemes


def run_pytest(
    *tests: Callable[..., Any],
    fixtures: Sequence[Callable[..., Any]] = (),
    expected_failed: int = 0,
) -> None:
    """
    Run the supplied test functions in-process using pytest, with the supplied fixtures,
    and assert that the expected number of tests ran and failed.

    :param tests: The test functions to run.
    :param fixtures: A fixture function, or sequence of fixture functions, needed by the tests.
    :param expected_failed: The number of tests expected to fail, defaulting to ``0``.
    """
    # local imports in case pytest is not available
    import pytest
    from _pytest.config import get_config
    from _pytest.main import Session
    from _pytest.runner import runtestprotocol

    # Build a virtual module containing our fixture and test functions.
    module = types.ModuleType(f'_sybil_run_pytest_{id(tests)}')
    for func in list(fixtures) + list(tests):
        setattr(module, func.__name__, func)

    # Create an isolated config using only pytest's built-in default plugins.
    # Patching load_setuptools_entrypoints on our own private PytestPluginManager
    # prevents external plugins (e.g. pytest-django) from being loaded and their
    # pytest_configure hooks from firing — without affecting the outer session.
    config = get_config(args=['-p', 'no:terminal'])
    config.pluginmanager.load_setuptools_entrypoints = lambda *a, **kw: None  # type: ignore[method-assign,assignment]
    config.parse(['--noconftest'], addopts=False)
    config.option.tbstyle = 'short'
    config._do_configure()

    class VirtualModule(pytest.Module):
        def _getobj(self) -> object:
            return module

    failures: list[str] = []
    session = Session.from_config(config)
    config.hook.pytest_sessionstart(session=session)

    virtual_module = VirtualModule.from_parent(session, path=Path(__file__))
    session._fixturemanager.parsefactories(virtual_module)

    items = [pytest.Function.from_parent(virtual_module, name=f.__name__, callobj=f) for f in tests]

    for i, item in enumerate(items):
        nextitem = items[i + 1] if i + 1 < len(items) else None
        reports = runtestprotocol(item, nextitem=nextitem)
        for report in reports:
            if report.failed and report.when in ('setup', 'call'):
                failures.append(str(report.longrepr))

    config.hook.pytest_sessionfinish(session=session, exitstatus=0)
    config._ensure_unconfigure()

    actual_failed = len(failures)
    if actual_failed != expected_failed:
        if actual_failed == 0:
            raise AssertionError(f'Expected {expected_failed} test(s) to fail, but none did')
        failure_text = '\n\n'.join(failures)
        raise AssertionError(
            f'Expected {expected_failed} test(s) to fail, but {actual_failed} did:\n\n'
            f'{failure_text}'
        )



import re

import doctest
from doctest import set_unittest_reportflags
from typing import Any, Dict, List, Optional

from sybil import Example

#: A doctest option flag to limit comparison of numbers in expected output to the precision
#: specified in the example.
NUMBER = doctest.register_optionflag("NUMBER")


class DocTest(doctest.DocTest):
    def __init__(
        self,
        examples: List[doctest.Example],
        globs: Dict[str, Any],
        name: str,
        filename: Optional[str],
        lineno: Optional[int],
        docstring: Optional[str],
    ) -> None:
        # do everything like regular doctests, but don't make a copy of globs
        doctest.DocTest.__init__(self, examples, globs, name, filename, lineno, docstring)
        self.globs = globs


def float_approx_equal(expected: str, actual: str, tolerance: float = 1e-12) -> bool:
    return abs(float(expected) - float(actual)) <= tolerance


class OutputChecker(doctest.OutputChecker):
    _number_re = re.compile(
        r"""
        (?P<number>
          (?P<mantissa>
            (?P<integer1> [+-]?\d*)\.(?P<fraction>\d+)
            |
            (?P<integer2> [+-]?\d+)\.
          )
          (?:
            [Ee]
            (?P<exponent1> [+-]?\d+)
          )?
          |
          (?P<integer3> [+-]?\d+)
          (?:
            [Ee]
            (?P<exponent2> [+-]?\d+)
          )
        )
        """,
        re.VERBOSE,
    )

    def _remove_unwanted_precision(self, want: str, got: str) -> str:
        wants = list(self._number_re.finditer(want))
        gots = list(self._number_re.finditer(got))
        if len(wants) != len(gots):
            return got
        offset = 0
        for w, g in zip(wants, gots):
            fraction: Optional[str] = w.group("fraction")
            exponent: Optional[str] = w.group("exponent1")
            if exponent is None:
                exponent = w.group("exponent2")
            if fraction is None:
                precision = 0
            else:
                precision = len(fraction)
            if exponent is not None:
                precision -= int(exponent)
            if float_approx_equal(w.group(), g.group(), tolerance=10**-precision):
                # They're close enough. Replace the text we actually
                # got with the text we want, so that it will match when we
                # check the string literally.
                got = got[: g.start() + offset] + w.group() + got[g.end() + offset :]
                offset += w.end() - w.start() - (g.end() - g.start())
        return got

    def check_output(self, want: str, got: str, optionflags: int) -> bool:
        allow_number = optionflags & NUMBER
        if allow_number:
            got = self._remove_unwanted_precision(want, got)
        return doctest.OutputChecker.check_output(self, want, got, optionflags)


class DocTestRunner(doctest.DocTestRunner):
    def __init__(self, optionflags: int) -> None:
        _unittest_reportflags = set_unittest_reportflags(0)
        set_unittest_reportflags(_unittest_reportflags)
        optionflags |= _unittest_reportflags
        doctest.DocTestRunner.__init__(
            self,
            verbose=False,
            optionflags=optionflags,
            checker=OutputChecker(),
        )

    def _failure_header(self, test: DocTest, example: doctest.Example) -> str:
        return ''


class DocTestEvaluator:
    """
    The :any:`Evaluator` to use for :class:`Regions <sybil.Region>` yielded by
    a :class:`~sybil.parsers.abstract.doctest.DocTestStringParser`.


    :param optionflags:
        :ref:`doctest option flags<option-flags-and-directives>` to use
        when evaluating examples.
    """

    def __init__(self, optionflags: int = 0) -> None:
        self.runner = DocTestRunner(optionflags)

    def __call__(self, sybil_example: Example) -> str:
        __tracebackhide__ = True
        example = sybil_example.parsed
        namespace = sybil_example.namespace
        output: List[str] = []
        remove_name = False
        try:
            if '__name__' not in namespace:
                remove_name = True
                namespace['__name__'] = '__test__'
            self.runner.run(
                DocTest(
                    [example],
                    namespace,
                    name=sybil_example.path,
                    filename=None,
                    lineno=example.lineno,
                    docstring=None,
                ),
                clear_globs=False,
                out=output.append,
            )
        finally:
            if remove_name:
                del namespace['__name__']
        return ''.join(output)
